'''
Migration support as features in MongoWorld 2016
From the Polls to the Trolls.

The Manager class provides a general purpose support
for migrating documents to a target version through
a series of migration functions.
'''

import re
import itertools

from typing import Set, Callable

from more_itertools import recipes


class Manager(object):
    """
    A manager for facilitating the registration of migration functions
    and applying those migrations to documents.

    To use, implement migration functions to and from each adjacent
    version of your schema and decorate each with the register
    classmethod. For example:

    >>> @Manager.register
    ... def v1_to_2(manager, doc):
    ...     doc['foo'] = 'bar'
    >>> @Manager.register
    ... def v2_to_3(manager, doc):
    ...     doc['foo'] = doc['foo'] + ' baz'

    Note that in addition to the document, the migration manager is also
    passed to the migration function, allowing for other context to be
    made available during the migration.

    To create a manager for migrating documents to version 3:

    >>> mgr = Manager(3)

    Then, use the manager to migrate a document to a target version.

    >>> v1_doc = dict(version=1, data='stub')
    >>> v3_doc = mgr.migrate_doc(v1_doc)

    >>> v3_doc['version']
    3
    >>> v3_doc['foo']
    'bar baz'

    Note that the document is modified in place:

    >>> v1_doc is v3_doc
    True

    >>> Manager._upgrade_funcs.clear()
    """

    version_attribute_name = 'version'
    _upgrade_funcs: Set[Callable] = set()

    def __init__(self, target_version):
        self.target_version = target_version

    @classmethod
    def register(cls, func):
        """
        Decorate a migration function with this method
        to make it available for migrating cases.
        """
        cls._add_version_info(func)
        cls._upgrade_funcs.add(func)
        return func

    @staticmethod
    def _add_version_info(func):
        """
        Add .source and .target attributes to the registered function.
        """
        pattern = r'v(?P<source>\d+)_to_(?P<target>\d+)$'
        match = re.match(pattern, func.__name__)
        if not match:
            raise ValueError("migration function name must match " + pattern)
        func.source, func.target = map(int, match.groups())

    def migrate_doc(self, doc):
        """
        Migrate the doc from its current version to the target version
        and return it.
        """
        orig_ver = doc.get(self.version_attribute_name, 0)
        funcs = self._get_migrate_funcs(orig_ver, self.target_version)
        for func in funcs:
            func(self, doc)
            doc[self.version_attribute_name] = func.target
        return doc

    @classmethod
    def _get_migrate_funcs(cls, orig_version, target_version):
        """
        >>> @Manager.register
        ... def v1_to_2(manager, doc):
        ...     doc['foo'] = 'bar'
        >>> @Manager.register
        ... def v2_to_1(manager, doc):
        ...     del doc['foo']
        >>> @Manager.register
        ... def v2_to_3(manager, doc):
        ...     doc['foo'] = doc['foo'] + ' baz'
        >>> funcs = list(Manager._get_migrate_funcs(1, 3))
        >>> len(funcs)
        2
        >>> funcs == [v1_to_2, v2_to_3]
        True
        >>> funcs = list(Manager._get_migrate_funcs(2, 1))
        >>> len(funcs)
        1
        >>> funcs == [v2_to_1]
        True

        >>> Manager._upgrade_funcs.clear()
        """
        direction = 1 if target_version > orig_version else -1
        versions = range(orig_version, target_version + direction, direction)
        transitions = recipes.pairwise(versions)
        return itertools.starmap(cls._get_func, transitions)

    @classmethod
    def _get_func(cls, source_ver, target_ver):
        """
        Return exactly one function to convert from source to target
        """
        matches = (
            func
            for func in cls._upgrade_funcs
            if func.source == source_ver and func.target == target_ver
        )
        try:
            (match,) = matches
        except ValueError:
            raise ValueError(f"No migration from {source_ver} to {target_ver}")
        return match



"""
Script to move a subset of GridFS files from one db
(or collection) to another.

>>> import io
>>> db_uri = getfixture('mongodb_uri') + '/move_gridfs_test'
>>> source = helper.connect_gridfs(db_uri + '.source')
>>> dest = helper.connect_gridfs(db_uri + '.dest')
>>> id = source.put(io.BytesIO(b'test'), filename='test.txt')
>>> mover = FileMove(source_gfs=source, dest_gfs=dest, delete=True)
>>> mover.ensure_indexes()
>>> mover.run(bar=None)
>>> source.list()
[]
>>> dest.exists(id)
True
>>> dest.list()
['test.txt']
"""

import sys
import logging
import argparse
import itertools
import signal

import bson
import dateutil.parser
from jaraco.ui import progress
from more_itertools.recipes import consume

from jaraco.mongodb import helper


log = logging.getLogger()


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('source_gfs', type=helper.connect_gridfs)
    parser.add_argument('dest_gfs', type=helper.connect_gridfs)
    parser.add_argument(
        '--include',
        help="a filter of files (regex) to include",
    )
    parser.add_argument(
        '--delete',
        default=False,
        action="store_true",
        help="delete files after moving",
    )
    parser.add_argument('--limit', type=int)
    parser.add_argument(
        '--limit-date',
        type=dateutil.parser.parse,
        help="only move files older than this date",
    )
    return parser.parse_args()


class FileMove:
    include = None
    delete = False
    limit = None
    limit_date = None

    def __init__(self, **params):
        vars(self).update(**params)

    def ensure_indexes(self):
        """
        Create the same indexes that the GridFS API would have
        """
        self.dest_gfs.new_file()._GridIn__ensure_indexes()

    @property
    def filter(self):
        filter = {}
        if self.include:
            filter.update(filename={"$regex": self.include})
        if self.limit_date:
            id_max = bson.objectid.ObjectId.from_datetime(self.limit_date)
            filter.update(_id={"$lt": id_max})
        return filter

    @property
    def source_coll(self):
        return self.source_gfs._GridFS__collection

    @property
    def dest_coll(self):
        return self.dest_gfs._GridFS__collection

    def run(self, bar=progress.TargetProgressBar):
        files = self.source_coll.files.find(
            self.filter,
            batch_size=1,
        )
        limit_files = itertools.islice(files, self.limit)
        count = min(files.count(), self.limit or float('inf'))
        progress = bar(count).iterate if bar else iter
        with SignalTrap(progress(limit_files)) as items:
            consume(map(self.process, items))

    def process(self, file):
        chunks = self.source_coll.chunks.find(dict(files_id=file['_id']))
        for chunk in chunks:
            self.dest_coll.chunks.insert(chunk)
        self.dest_coll.files.insert(file)
        self.delete and self.source_gfs.delete(file['_id'])


class SignalTrap:
    """
    A context manager for wrapping an iterable such that it
    is only interrupted between iterations.
    """

    def __init__(self, iterable):
        self.iterable = iterable

    def __enter__(self):
        self.prev = signal.signal(signal.SIGINT, self.stop)
        return self

    def __exit__(self, *args):
        signal.signal(signal.SIGINT, self.prev)
        del self.prev

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.iterable)

    next = __next__

    def stop(self, signal, frame):
        self.iterable = iter([])


def run():
    logging.basicConfig(stream=sys.stderr, level=logging.INFO)
    args = get_args()

    mover = FileMove(**vars(args))
    mover.ensure_indexes()
    mover.run()


if __name__ == '__main__':
    run()



import shlex
import contextlib

import pmxbot.storage

from . import sharding


class Storage(pmxbot.storage.SelectableStorage, pmxbot.storage.MongoDBStorage):
    collection_name = 'unused'


def get_client():
    """
    Use the same MongoDB client as pmxbot if available.
    """
    with contextlib.suppress(Exception):
        store = Storage.from_URI()
        assert isinstance(store, pmxbot.storage.MongoDBStorage)
        return store.db.database.client


@pmxbot.core.command("create-db-in-shard")
def cdbs(client, event, channel, nick, rest):
    """
    Create a database in a shard. !create-db-in-shard {db name} {shard name}
    """
    db_name, shard = shlex.split(rest)
    return sharding.create_db_in_shard(db_name, shard, client=get_client())



import builtins


def estimate(coll, filter={}, sample=1):
    """
    Estimate the number of documents in the collection
    matching the filter.

    Sample may be a fixed number of documents to sample
    or a percentage of the total collection size.

    >>> coll = getfixture('bulky_collection')
    >>> estimate(coll)
    100
    >>> query = {"val": {"$gte": 50}}
    >>> val = estimate(coll, filter=query)
    >>> val > 0
    True
    >>> val = estimate(coll, filter=query, sample=10)
    >>> val > 0
    True
    >>> val = estimate(coll, filter=query, sample=.1)
    >>> val > 0
    True
    """
    total = coll.estimated_document_count()
    if not filter and sample == 1:
        return total
    if sample <= 1:
        sample *= total
    pipeline = list(
        builtins.filter(
            None,
            [
                {'$sample': {'size': sample}} if sample < total else {},
                {'$match': filter},
                {'$count': 'matched'},
            ],
        )
    )
    docs = next(coll.aggregate(pipeline))
    ratio = docs['matched'] / sample
    return int(total * ratio)



"""
Script to check a GridFS instance for corrupted records.
"""

import sys
import logging
import argparse

import pymongo
from jaraco.ui import progress
from more_itertools.recipes import consume
from jaraco.itertools import Counter

from jaraco.mongodb import helper
from jaraco.context import ExceptionTrap


log = logging.getLogger()


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--depth',
        default=1024,
        help="Bytes to read into each file during check",
    )
    parser.add_argument('db', type=helper.connect_gridfs)
    return parser.parse_args()


class FileChecker:
    def __init__(self, gfs, depth):
        self.gfs = gfs
        self.depth = depth

    def run(self):
        files = self.gfs.list()
        bar = progress.TargetProgressBar(len(files))
        processed_files = map(self.process, bar.iterate(files))
        errors = filter(None, processed_files)
        counter = Counter(errors)
        consume(map(self.handle_trap, counter))
        return counter

    def process(self, filename):
        file = self.gfs.get_last_version(filename)
        with ExceptionTrap(pymongo.errors.PyMongoError) as trap:
            file.read(self.depth)
        trap.filename = filename
        return trap

    def handle_trap(self, trap):
        cls, exc, tb = trap.exc_info
        log.error("Failed to read %s (%s)", trap.filename, exc)


def run():
    logging.basicConfig(stream=sys.stderr)
    args = get_args()

    checker = FileChecker(args.db, args.depth)
    counter = checker.run()

    print("Encountered", counter.count, "errors")


if __name__ == '__main__':
    run()



from jaraco.mongodb import compat


def test_save_no_id(database):
    doc = dict(foo='bar')
    compat.save(database.test_coll, doc)
    assert database.test_coll.find_one()['foo'] == 'bar'


def test_save_new_with_id(database):
    doc = dict(foo='bar', _id=1)
    compat.save(database.test_coll, doc)
    assert database.test_coll.find_one() == doc


def test_save_replace_by_id(database):
    compat.save(database.test_coll, dict(foo='bar', _id=1))

    doc = dict(foo='baz', _id=1)
    compat.save(database.test_coll, doc)
    assert database.test_coll.count_documents({}) == 1
    assert database.test_coll.find_one() == doc


def test_save_no_id_extant_docs(database):
    """
    When no id is supplied, a new document should be created.
    """
    doc = dict(foo='bar')
    compat.save(database.test_coll, dict(doc))
    assert database.test_coll.count_documents({}) == 1
    compat.save(database.test_coll, doc)
    assert database.test_coll.count_documents({}) == 2


def test_save_adds_id(database):
    """
    Ensure _id is added to an inserted document.
    """
    doc = dict(foo='bar')
    compat.save(database.test_coll, doc)
    assert '_id' in doc



import pprint
import textwrap

from .query import compat_explain


def _rep_index_info(coll):
    index_info = coll.index_information()
    return "Indexes are:\n" + pprint.pformat(index_info)


def assert_index_used(cur):
    """
    Explain the cursor and ensure that the index was used.
    """
    explanation = compat_explain(cur)
    plan = explanation['queryPlanner']['winningPlan']
    assert plan['stage'] != 'COLLSCAN'
    assert plan['inputStage']['stage'] == 'IXSCAN'


def assert_covered(cur):
    """
    Use the best knowledge about Cursor.explain() to ensure that the query
    was covered by an index.
    """
    explanation = compat_explain(cur)
    tmpl = textwrap.dedent(
        """
        Query was not covered:
        {explanation}
        """
    ).lstrip()
    report = tmpl.format(explanation=pprint.pformat(explanation))
    report += _rep_index_info(cur.collection)
    stats = explanation['executionStats']
    assert stats['totalDocsExamined'] == 0, report
    assert stats['totalKeysExamined'], "No documents matched"
    return explanation


def assert_distinct_covered(coll, field, query):
    """
    Ensure a distinct query is covered by an index.
    """
    est = coll.estimated_document_count()
    assert est, "Unable to assert without a document"
    db = coll.database
    res = db.command('distinct', coll.name, key=field, query=query)
    assert 'stats' in res, "Stats not supplied. Maybe SERVER-9126?"
    stats = res['stats']
    tmpl = textwrap.dedent(
        """
        Distinct query was not covered:
        {explanation}
        """
    ).lstrip()
    report = tmpl.format(explanation=pprint.pformat(stats))
    report += _rep_index_info(coll)
    assert stats['nscannedObjects'] == 0, report
    assert stats['n'], "No documents matched"
    return stats



"""
Decode support for JSON strings using ordered dictionaries and parsing
MongoDB-specific objects like dates.

For example, if you have a JSON object representing a sort, you
want to retain the order of keys:

>>> ob = decode('{"key1": 1, "key2": 2}')
>>> list(ob.keys())
['key1', 'key2']

Or if you want to query by a date, PyMongo needs a Python datetime
object, which has no JSON representation, so this codec converts
``$date`` keys to date objects.

>>> ob = decode('{"$gte": {"$date": "2019-01-01"}}')
>>> ob['$gte']
datetime.datetime(2019, 1, 1, 0, 0)

This function is useful in particular if you're accepting JSON queries
over an HTTP connection and you don't have the luxury of Javascript
expressions like you see in the Mongo shell or Compass.
"""

import json
import functools
import collections

import dateutil.parser
from jaraco.functools import compose


def maybe_date(obj):
    """
    >>> maybe_date({"$date": "2019-01-01"})
    datetime.datetime(2019, 1, 1, 0, 0)
    """
    return dateutil.parser.parse(obj['$date']) if list(obj) == ['$date'] else obj


smart_hook = compose(maybe_date, collections.OrderedDict)


decode = functools.partial(json.loads, object_pairs_hook=smart_hook)



import urllib.parse

import jaraco.functools


@jaraco.functools.once
def _add_scheme():
    """
    urllib.parse doesn't support the mongodb scheme, but it's easy
    to make it so.
    """
    lists = [
        urllib.parse.uses_relative,
        urllib.parse.uses_netloc,
        urllib.parse.uses_query,
    ]
    for each in lists:
        each.append('mongodb')


def join(base, new):
    """
    Use urllib.parse to join the MongoDB URIs.
    Registers the MongoDB scheme first.
    """
    _add_scheme()
    return urllib.parse.urljoin(base, new)



import argparse
import time
import json
import logging
import pymongo
import bson.json_util
import re
import collections
import datetime

try:
    from importlib import metadata  # type: ignore
except ImportError:
    import importlib_metadata as metadata  # type: ignore

from typing import Dict, Any

import jaraco.logging
import pytimeparse
from jaraco.functools import compose
from pymongo.cursor import CursorType
from jaraco.itertools import always_iterable
from jaraco.ui.cmdline import Extend


def delta_from_seconds(seconds):
    return datetime.timedelta(seconds=int(seconds))


def parse_args(*args, **kwargs):
    """
    Parse the args for the command.

    It should be possible for one to specify '--ns', '-x', and '--rename'
    multiple times:

    >>> args = parse_args(['--ns', 'foo', 'bar', '--ns', 'baz'])
    >>> args.ns
    ['foo', 'bar', 'baz']

    >>> parse_args(['-x', '--exclude']).exclude
    []

    >>> renames = parse_args(['--rename', 'a=b', '--rename', 'b=c']).rename
    >>> len(renames)
    2

    "..." below should be "jaraco." but for pytest-dev/pytest#3396
    >>> type(renames)
    <class '...mongodb.oplog.Renamer'>
    """
    parser = argparse.ArgumentParser(add_help=False)

    parser.add_argument(
        "--help",
        help="show usage information",
        action="help",
    )

    parser.add_argument(
        "--source",
        metavar="host[:port]",
        help="""Hostname of the mongod server from which oplog
        operations are going to be pulled. Called "--from"
        in mongooplog.""",
    )

    parser.add_argument(
        '--oplogns',
        default='local.oplog.rs',
        help="Source namespace for oplog",
    )

    parser.add_argument(
        "--dest",
        metavar="host[:port]",
        help="""
        Hostname of the mongod server (or replica set as
        <set name>/s1,s2) to which oplog operations
        are going to be applied. Default is "localhost".
        Called "--host" in mongooplog.
        """,
    )

    parser.add_argument(
        "-w",
        "--window",
        dest="start_ts",
        metavar="WINDOW",
        type=compose(
            Timestamp.for_window,
            delta_from_seconds,
            pytimeparse.parse,
        ),
        help="""Time window to query, like "3 days" or "24:00"
        (24 hours, 0 minutes).""",
    )

    parser.add_argument(
        "-f",
        "--follow",
        action="store_true",
        help="""Wait for new data in oplog. Makes the utility
        polling oplog forever (until interrupted). New data
        is going to be applied immediately with at most one
        second delay.""",
    )

    parser.add_argument(
        "--ns",
        nargs="*",
        default=[],
        action=Extend,
        help="""Process only these namespaces, ignoring all others.
        Space separated list of strings in form of ``dname``
        or ``dbname.collection``. May be specified multiple times.
        """,
    )

    parser.add_argument(
        "-x",
        "--exclude",
        nargs="*",
        default=[],
        action=Extend,
        help="""List of space separated namespaces which should be
        ignored. Can be in form of ``dname`` or ``dbname.collection``.
        May be specified multiple times.
        """,
    )

    parser.add_argument(
        "--rename",
        nargs="*",
        default=[],
        metavar="ns_old=ns_new",
        type=RenameSpec.from_spec,
        action=Extend,
        help="""
        Rename database(s) and/or collection(s). Operations on
        namespace ``ns_old`` from the source server will be
        applied to namespace ``ns_new`` on the destination server.
        May be specified multiple times.
        """,
    )

    parser.add_argument(
        "--dry-run",
        default=False,
        action="store_true",
        help="Suppress application of ops.",
    )

    parser.add_argument(
        "--resume-file",
        metavar="FILENAME",
        type=ResumeFile,
        default=NullResumeFile(),
        help="""Read from and write to this file the last processed
        timestamp.""",
    )

    jaraco.logging.add_arguments(parser)

    args = parser.parse_args(*args, **kwargs)
    args.rename = Renamer(args.rename)

    args.start_ts = args.start_ts or args.resume_file.read()

    return args


class RenameSpec(object):
    @classmethod
    def from_spec(cls, string_spec):
        """
        Construct RenameSpec from a pair separated by equal sign ('=').
        """
        old_ns, new_ns = string_spec.split('=')
        return cls(old_ns, new_ns)

    def __init__(self, old_ns, new_ns):
        self.old_ns = old_ns
        self.new_ns = new_ns
        self.old_db, sep, self.old_coll = self.old_ns.partition('.')
        self.new_db, sep, self.new_coll = self.new_ns.partition('.')
        self.regex = re.compile(r"^{0}(\.|$)".format(re.escape(self.old_ns)))

        # ugly hack: append a period so the regex can match dot
        # or end of string; requires .rstrip operation in __call__ also.
        self.new_ns += "."

    def __call__(self, op):
        """
        Apply this rename to the op
        """
        self._handle_renameCollection(op)
        if self.regex.match(op['ns']):
            ns = self.regex.sub(self.new_ns, op['ns']).rstrip(".")
            logging.debug("renaming %s to %s", op['ns'], ns)
            op['ns'] = ns
        if op['ns'].endswith('.system.indexes'):
            # index operation; update ns in the op also.
            self(op['o'])
        self._handle_create(op)

    def _handle_create(self, op):
        # MongoDB 3.4 introduces idIndex
        if 'idIndex' in op.get('o', {}):
            self(op['o']['idIndex'])
        if self._matching_create_command(op, self.old_ns):
            op['ns'] = self.new_db + '.$cmd'
            op['o']['create'] = self.new_coll

    @staticmethod
    def _matching_create_command(op, ns):
        db, sep, coll = ns.partition('.')
        return (
            op.get('op') == 'c'
            and op['ns'] == db + '.$cmd'
            and op['o'].get('create', None) == coll
        )

    @staticmethod
    def _matching_renameCollection_command(op, ns):
        db, sep, coll = ns.partition('.')
        return (
            op.get('op') == 'c'
            and (
                # seems command can happen in admin or the db
                op['ns'] == 'admin.$cmd'
                or op['ns'] == db + '.$cmd'
            )
            and 'renameCollection' in op['o']
            and (
                op['o']['renameCollection'].startswith(ns)
                or op['o']['to'].startswith(ns)
            )
        )

    def _handle_renameCollection(self, op):
        if self._matching_renameCollection_command(op, self.old_ns):
            cmd = op['o']
            rename_keys = 'renameCollection', 'to'
            for key in rename_keys:
                # todo, this is a mirror of the code in __call__; refactor
                if self.regex.match(cmd[key]):
                    ns = self.regex.sub(self.new_ns, cmd[key]).rstrip(".")
                    logging.debug("renaming %s to %s", cmd[key], ns)
                    cmd[key] = ns

    def affects(self, ns):
        return bool(self.regex.match(ns))


class Renamer(list):
    """
    >>> specs = [
    ...      'a=b',
    ...      'alpha=gamma',
    ...  ]
    >>> renames = Renamer.from_specs(specs)
    >>> op = dict(ns='a.a')
    >>> renames(op)
    >>> op['ns']
    'b.a'
    >>> renames.affects('alpha.foo')
    True
    >>> renames.affects('b.gamma')
    False
    """

    def invoke(self, op):
        """
        Replace namespaces in op based on RenameSpecs in self.
        """
        for rename in self:
            rename(op)

    __call__ = invoke

    @classmethod
    def from_specs(cls, specs):
        return cls(map(RenameSpec.from_spec, always_iterable(specs)))

    def affects(self, ns):
        """
        Return True if this renamer affects the indicated namespace.
        """
        return any(rn.affects(ns) for rn in self)


def string_none(value):
    """
    Convert the string 'none' to None
    """
    is_string_none = not value or value.lower() == 'none'
    return None if is_string_none else value


def _same_instance(client1, client2):
    """
    Return True if client1 and client2 appear to reference the same
    MongoDB instance.
    """
    return client1._topology_settings.seeds == client2._topology_settings.seeds


def _full_rename(args):
    """
    Return True only if the arguments passed specify exact namespaces
    and to conduct a rename of every namespace.
    """
    return args.ns and all(map(args.rename.affects, args.ns))


def _resolve_shard(client):
    """
    The destination cannot be a mongoS instance, as applyOps is
    not an allowable command for mongoS instances, so if the
    client is a connection to a mongoS instance, raise an error
    or resolve the replica set.
    """
    status = client.admin.command('serverStatus')
    if status['process'] == 'mongos':
        raise RuntimeError("Destination cannot be mongos")
    return client


def _load_dest(host):
    if not host:
        return
    return _resolve_shard(pymongo.MongoClient(host))


def main():
    args = parse_args()
    log_format = '%(asctime)s - %(levelname)s - %(message)s'
    jaraco.logging.setup(args, format=log_format)

    logging.info(f"jaraco.mongodb.oplog {metadata.version('jaraco.mongodb')}")
    logging.info("going to connect")

    src = pymongo.MongoClient(args.source)
    dest = _load_dest(args.dest)

    if dest and _same_instance(src, dest) and not _full_rename(args):
        logging.error(
            "source and destination hosts can be the same only "
            "when both --ns and --rename arguments are given"
        )
        raise SystemExit(1)

    logging.info("connected")

    start = args.start_ts
    if not start:
        logging.error("Resume file or window required")
        raise SystemExit(2)

    logging.info("starting from %s (%s)", start, start.as_datetime())
    db_name, sep, coll_name = args.oplogns.partition('.')
    oplog_coll = src[db_name][coll_name]
    num = 0

    class_ = TailingOplog if args.follow else Oplog
    generator = class_(oplog_coll)

    if not generator.has_ops_before(start):
        logging.warning("No ops before start time; oplog may be overrun")

    try:
        for num, doc in enumerate(generator.since(start)):
            _handle(dest, doc, args, num)
            last_handled = doc
        logging.info("all done")
    except KeyboardInterrupt:
        logging.info("Got Ctrl+C, exiting...")
    finally:
        if 'last_handled' in locals():
            last = last_handled['ts']
            args.resume_file.save(last)
            logging.info("last ts was %s (%s)", last, last.as_datetime())


def applies_to_ns(op, ns):
    return (
        op['ns'].startswith(ns)
        or RenameSpec._matching_create_command(op, ns)
        or RenameSpec._matching_renameCollection_command(op, ns)
    )


class NiceRepr(object):
    """
    Adapt a Python representation of a MongoDB object
    to make it appear nicely when rendered as a
    string.

    >>> messy_doc = collections.OrderedDict([
    ...     ('ts', bson.Timestamp(1111111111, 30),
    ... )])
    >>> print(NiceRepr(messy_doc))
    {"ts": {"$timestamp": {"t": 1111111111, "i": 30}}}
    """

    def __init__(self, orig):
        self.__orig = orig

    def __str__(self):
        return bson.json_util.dumps(self.__orig)

    def __getattr__(self, attr):
        return getattr(self.__orig, attr)


def _handle(dest, op, args, num):
    # Skip "no operation" items
    if op['op'] == 'n':
        return

    # Skip excluded namespaces or namespaces that does not match --ns
    excluded = any(applies_to_ns(op, ns) for ns in args.exclude)
    included = any(applies_to_ns(op, ns) for ns in args.ns)

    if excluded or (args.ns and not included):
        logging.log(logging.DEBUG - 1, "skipping %s", op)
        return

    args.rename(op)

    logging.debug("applying op %s", NiceRepr(op))
    try:
        args.dry_run or apply(dest, op)
    except pymongo.errors.OperationFailure as e:
        nice_op = NiceRepr(op)
        msg = f'{e!r} applying {nice_op}'
        logging.warning(msg)

    # Update status
    ts = op['ts']
    if not num % 1000:
        args.resume_file.save(ts)
        logging.info(
            "%s\t%s\t%s -> %s",
            num,
            ts.as_datetime(),
            op.get('op'),
            op.get('ns'),
        )


def apply(db, op):
    """
    Apply operation in db
    """
    dbname = op['ns'].split('.')[0] or "admin"
    _db = db[dbname]
    return _apply_index_op(_db, op) or _apply_regular(_db, op)


def _apply_index_op(db, op):
    """
    Starting with MongoDB 4.2, index operations can no longer
    be applied. Intercept the application and transform it into
    a normal create index operation.
    """
    if 'createIndexes' not in op['o']:
        return
    o = op['o']
    coll_name = o['createIndexes']
    key = list(o['key'].items())
    name = o['name']
    return db[coll_name].create_index(key, name=name)


def _apply_regular(db, op):
    opts = bson.CodecOptions(uuid_representation=bson.binary.STANDARD)
    db.command("applyOps", [op], codec_options=opts)


class Oplog(object):
    find_params: Dict[str, Any] = {}

    def __init__(self, coll):
        self.coll = coll.with_options(
            codec_options=bson.CodecOptions(
                document_class=collections.OrderedDict,
            ),
        )

    def get_latest_ts(self):
        cur = self.coll.find().sort('$natural', pymongo.DESCENDING).limit(-1)
        latest_doc = next(cur)
        return Timestamp.wrap(latest_doc['ts'])

    def query(self, spec):
        return self.coll.find(spec, **self.find_params)

    def since(self, ts):
        """
        Query the oplog for items since ts and then return
        """
        spec = {'ts': {'$gt': ts}}
        cursor = self.query(spec)
        while True:
            # todo: trap InvalidDocument errors:
            # except bson.errors.InvalidDocument as e:
            #  logging.info(repr(e))
            for doc in cursor:
                yield doc
            if not cursor.alive:
                break
            time.sleep(1)

    def has_ops_before(self, ts):
        """
        Determine if there are any ops before ts
        """
        spec = {'ts': {'$lt': ts}}
        return bool(self.coll.find_one(spec))


class TailingOplog(Oplog):
    find_params = dict(
        cursor_type=CursorType.TAILABLE_AWAIT,
        oplog_replay=True,
    )

    def since(self, ts):
        """
        Tail the oplog, starting from ts.
        """
        while True:
            items = super(TailingOplog, self).since(ts)
            for doc in items:
                yield doc
                ts = doc['ts']


class Timestamp(bson.timestamp.Timestamp):
    @classmethod
    def wrap(cls, orig):
        """
        Wrap an original timestamp as returned by a pymongo query
        with a version of this class.
        """
        # hack to give the timestamp this class' specialized methods
        orig.__class__ = cls
        return orig

    def dump(self, stream):
        """Serialize self to text stream.

        Matches convention of mongooplog.
        """
        items = (
            ('time', self.time),
            ('inc', self.inc),
        )
        # use ordered dict to retain order
        ts = collections.OrderedDict(items)
        json.dump(dict(ts=ts), stream)

    @classmethod
    def load(cls, stream):
        """Load a serialized version of self from text stream.

        Expects the format used by mongooplog.
        """
        data = json.load(stream)['ts']
        return cls(data['time'], data['inc'])

    @classmethod
    def for_window(cls, window):
        """
        Given a timedelta window, return a timestamp representing
        that time.
        """
        utcnow = datetime.datetime.utcnow()
        return cls(utcnow - window, 0)


class ResumeFile(str):
    def save(self, ts):
        """
        Save timestamp to file.
        """
        with open(self, 'w') as f:
            Timestamp.wrap(ts).dump(f)

    def read(self):
        """
        Read timestamp from file.
        """
        with open(self) as f:
            return Timestamp.load(f)


class NullResumeFile(object):
    def save(self, ts):
        pass

    def read(self):
        pass


if __name__ == '__main__':
    main()



import datetime

import pymongo
import pytest
from tempora import timing

from jaraco.mongodb import service


@pytest.mark.xfail(reason="#31")
def test_MongoDBReplicaSet_writable():
    rs = service.MongoDBReplicaSet()
    rs.start()
    try:
        conn = pymongo.MongoClient(rs.get_connect_hosts())
        conn.database.collection.insert_one({'foo': 'bar'})
    finally:
        rs.stop()


def test_MongoDBReplicaSet_starts_quickly():
    pytest.skip("Takes 20-30 seconds")
    sw = timing.Stopwatch()
    rs = service.MongoDBReplicaSet()
    rs.start()
    try:
        elapsed = sw.split()
        limit = datetime.timedelta(seconds=5)
        assert elapsed < limit
    finally:
        rs.stop()


def test_fixture(mongodb_instance):
    "Cause the fixture to be invoked"
    pass



import json
import pathlib
import platform
import re
import tarfile
import urllib.request
import posixpath
import zipfile
import io

import autocommand
from more_itertools import one


def get_download_url():
    source = 'https://www.mongodb.com/try/download/community'
    with urllib.request.urlopen(source) as resp:
        html = resp.read().decode('utf-8')
    server_data = re.search(
        r'<script id="server-data">\s*window\.__serverData=(.*?)\s*</script>',
        html,
        flags=re.DOTALL | re.MULTILINE,
    ).group(1)
    data = json.loads(server_data)
    versions = data['components'][2]['props']['embeddedComponents'][0]['props'][
        'items'
    ][2]['embeddedComponents'][0]['props']['data'][0]['data'][0]
    best_version = next(ver for ver in versions if versions[ver]['meta']['current'])
    platforms = versions[best_version]['platforms']
    lookup = {
        ('Darwin', 'arm64'): 'macOS ARM 64',
        ('Darwin', 'x86_64'): 'macOS x64',
        ('Linux', 'x86_64'): 'Ubuntu 22.04 x64',
        ('Linux', 'aarch64'): 'Ubuntu 22.04 ARM 64',
        ('Windows', 'x86_64'): 'Windows x64',
        ('Windows', 'ARM64'): 'Windows x64',
    }
    plat_name = lookup[(platform.system(), platform.machine())]
    format = 'zip' if 'Windows' in plat_name else 'tgz'
    return platforms[plat_name][format]


class RootFinder(set):
    def __call__(self, info, path):
        self.add(self.root(info.name))
        return info

    @staticmethod
    def root(name):
        root, _, _ = name.partition(posixpath.sep)
        return root

    @classmethod
    def from_names(cls, names):
        return cls(map(cls.root, names))


def _extract_all(resp, target):
    desig = resp.headers['Content-Type'].lower().replace('/', '_').replace('+', '_')
    func_name = f'_extract_{desig}'
    return globals()[func_name](resp, target)


def _extract_application_zip(resp, target):
    data = io.BytesIO(resp.read())
    with zipfile.ZipFile(data) as obj:
        roots = RootFinder.from_names(obj.namelist())
        obj.extractall(
            target,
        )
    return roots


def _extract_application_gzip(resp, target):
    with tarfile.open(fileobj=resp, mode='r|*') as obj:
        roots = RootFinder()
        # python/typeshed#10514
        obj.extractall(target, filter=roots)  # type: ignore
    return roots


def install(target: pathlib.Path = pathlib.Path()):
    url = get_download_url()
    with urllib.request.urlopen(url) as resp:
        roots = _extract_all(resp, target.expanduser())
    return target.joinpath(one(roots))


autocommand.autocommand(__name__)(install)



import argparse


def extract_param(param, args, type=None):
    """
    From a list of args, extract the one param if supplied,
    returning the value and unused args.

    >>> extract_param('port', ['foo', '--port=999', 'bar'], type=int)
    (999, ['foo', 'bar'])
    >>> extract_param('port', ['foo', '--port', '999', 'bar'], type=int)
    (999, ['foo', 'bar'])
    >>> extract_param('port', ['foo', 'bar'])
    (None, ['foo', 'bar'])
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('--' + param, type=type)
    res, unused = parser.parse_known_args(args)
    return getattr(res, param), unused



import sys
import subprocess
import json


def test_insert_doc_command(mongodb_instance):
    uri = mongodb_instance.get_uri() + '/testdb.test_coll'
    cmd = [
        sys.executable,
        '-m',
        'jaraco.mongodb.insert-doc',
        uri,
    ]
    proc = subprocess.Popen(cmd, stdin=subprocess.PIPE)
    doc = dict(test='value', test2=2)
    proc.communicate(json.dumps(doc).encode('utf-8'))
    assert not proc.wait()
    (saved,) = mongodb_instance.get_connection().testdb.test_coll.find()
    saved.pop('_id')
    assert saved == doc



"""
Script to repair broken GridFS files. It handles

- Removing files with missing chunks.
"""

import sys
import logging
import argparse

import gridfs
from jaraco.ui import progress
from more_itertools.recipes import consume
from jaraco.itertools import Counter

from jaraco.mongodb import helper
from jaraco.context import ExceptionTrap


log = logging.getLogger()


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('db', type=helper.connect_gridfs)
    return parser.parse_args()


class FileRepair:
    def __init__(self, gfs):
        self.gfs = gfs
        db = gfs._GridFS__database
        coll = gfs._GridFS__collection
        bu_coll_name = coll.name + '-saved'
        self.backup_coll = db[bu_coll_name]

    def run(self):
        files = self.gfs.list()
        bar = progress.TargetProgressBar(len(files))
        processed_files = map(self.process, bar.iterate(files))
        errors = filter(None, processed_files)
        counter = Counter(errors)
        consume(map(self.handle_trap, counter))
        return counter

    def process(self, filename):
        file = self.gfs.get_last_version(filename)
        with ExceptionTrap(gridfs.errors.CorruptGridFile) as trap:
            file.read(1)
        trap.filename = filename
        return trap

    def handle_trap(self, trap):
        cls, exc, tb = trap.exc_info
        spec = dict(filename=trap.filename)
        for file_doc in self.gfs._GridFS__files.find(spec):
            self.backup_coll.files.insert(file_doc)
            chunk_spec = dict(files_id=file_doc['_id'])
            for chunk in self.gfs._GridFS__chunks.find(chunk_spec):
                self.backup_coll.chunks.insert(chunk)
        log.info("Removing %s (%s)", trap.filename, exc)
        self.gfs.delete(spec)


def run():
    logging.basicConfig(stream=sys.stderr, level=logging.INFO)
    args = get_args()

    repair = FileRepair(args.db)
    counter = repair.run()

    log.info("Removed %s corrupt files.", counter.count)


if __name__ == '__main__':
    run()



import functools

import pytest

from jaraco.mongodb import testing


@pytest.fixture
def indexed_collection(request, mongodb_instance):
    _name = request.function.__name__
    db = mongodb_instance.get_connection()[_name]
    dropper = functools.partial(db.client.drop_database, _name)
    request.addfinalizer(dropper)
    coll = db[_name]
    coll.create_index('foo')
    return coll


def test_assert_covered_passes(indexed_collection):
    indexed_collection.insert_one({'foo': 'bar'})
    proj = {'_id': False, 'foo': True}
    cur = indexed_collection.find({'foo': 'bar'}, proj)
    testing.assert_covered(cur)


def test_assert_covered_empty(indexed_collection):
    """
    assert_covered should raise an error it's trivially
    covered (returns no results)
    """
    cur = indexed_collection.find()
    with pytest.raises(AssertionError):
        testing.assert_covered(cur)


def test_assert_covered_null(indexed_collection):
    """
    assert_covered should raise an error it's trivially
    covered (returns no results)
    """
    indexed_collection.insert_one({"foo": "bar"})
    proj = {'_id': False, 'foo': True}
    cur = indexed_collection.find({"foo": "baz"}, proj)
    with pytest.raises(AssertionError):
        testing.assert_covered(cur)


def test_assert_index_used_passes(indexed_collection):
    """
    assert_index_used should pass when the index is used,
    even if the documents had to be hit.
    """
    indexed_collection.insert_one({'foo': 'bar', 'bing': 'baz'})
    cur = indexed_collection.find({'foo': 'bar'})
    testing.assert_index_used(cur)


def test_assert_index_used_fails(indexed_collection):
    """
    assert_index_used should fail when no index is used.
    """
    indexed_collection.insert_one({'foo': 'bar', 'bing': 'baz'})
    cur = indexed_collection.find({'bing': 'baz'})
    with pytest.raises(AssertionError):
        testing.assert_index_used(cur)



import shlex
import os

import pytest

try:
    import pymongo
except ImportError:
    pass

from . import service


def pytest_addoption(parser):
    parser.addoption(
        '--mongod-args',
        help="Arbitrary arguments to mongod",
    )
    parser.addoption(
        '--mongodb-uri',
        help="URI to an extant MongoDB instance (supersedes ephemeral)",
    )


@pytest.fixture(scope='session')
def mongodb_instance(request):
    if 'pymongo' not in globals():
        pytest.skip("pymongo not available")

    yield from _extant_instance(request.config)
    yield from _ephemeral_instance(request.config)


def _extant_instance(config):
    uri = config.getoption('mongodb_uri') or os.environ.get('MONGODB_URL')
    if not uri:
        return
    yield service.ExtantInstance(uri)


def _ephemeral_instance(config):
    params_raw = config.getoption('mongod_args') or ''
    params = shlex.split(params_raw)
    try:
        instance = service.MongoDBInstance()
        with instance.ensure():
            instance.merge_mongod_args(params)
            instance.start()
            pymongo.MongoClient(instance.get_connect_hosts())
            yield instance
    except Exception as err:
        raise
        pytest.skip(f"MongoDB not available ({err})")
    instance.stop()


@pytest.fixture(scope='session')
def mongodb_uri(mongodb_instance):
    return mongodb_instance.get_uri()



import sys
import json

import argparse

import pymongo.uri_parser


def get_collection(uri):
    parsed = pymongo.uri_parser.parse_uri(uri)
    client = pymongo.MongoClient(uri)
    return client[parsed['database']][parsed['collection']]


def parse_args():
    parser = argparse.ArgumentParser(
        "Insert a document from stdin into the specied collection"
    )
    parser.add_argument(
        'collection',
        metavar='collection_uri',
        type=get_collection,
    )
    return parser.parse_args()


def main():
    args = parse_args()
    args.collection.insert_one(json.load(sys.stdin))


__name__ == '__main__' and main()



from jaraco.mongodb import fields


def test_insert_with_dots(mongodb_instance):
    db = mongodb_instance.get_connection().test_db
    field = fields.encode("foo.bar")
    db.things.insert_one({field: "value"})
    doc = db.things.find_one({field: "value"})
    doc = {fields.decode(key): value for key, value in doc.items()}
    assert doc['foo.bar'] == 'value'



import functools

import pymongo.collection
from jaraco.collections import Projection


def save(coll, to_save):
    """
    Pymongo has deprecated the save logic, even though
    MongoDB still advertizes that logic in the core API:
    https://docs.mongodb.com/manual/reference/method/db.collection.save/

    This function provides a compatible interface.
    """
    filter = Projection(['_id'], to_save)
    upsert_replace = functools.partial(coll.replace_one, filter, upsert=True)
    op = upsert_replace if filter else coll.insert_one
    return op(to_save)


class Collection(pymongo.collection.Collection):
    """
    Subclass of default Collection that provides a non-deprecated
    save method. Don't use without first reading the cautions at
    https://github.com/mongodb/specifications/blob/master/source/crud/crud.rst#q--a

    >> db = getfixture('database')
    >> coll = Collection(db, 'mycoll')
    >> coll.save({'foo': 'bar'})
    <pymongo.results.InsertOneResult object at ...>
    >> ob = coll.find_one()
    >> ob['foo'] = 'baz'
    >> coll.save(ob)
    <pymongo.results.UpdateResult object at ...>
    """

    save = save


def query_or_command(op):
    """
    Given an operation from currentOp, return the query or command
    field as it changed in MongoDB 3.2 for indexing operations:

    https://docs.mongodb.com/manual/reference/method/db.currentOp/#active-indexing-operations
    """
    return op.get('command') or op.get('query')



import socket
import operator

import pymongo

hostname = socket.gethostname()
by_id = operator.itemgetter('_id')


def get_ids(collection):
    return map(by_id, collection.find(projection=['_id']))


def create_db_in_shard(db_name, shard, client=None):
    """
    In a sharded cluster, create a database in a particular shard.
    """
    client = client or pymongo.MongoClient()
    # flush the router config to ensure it's not stale
    res = client.admin.command('flushRouterConfig')
    if not res.get('ok'):
        raise RuntimeError("unable to flush router config")
    if shard not in get_ids(client.config.shards):
        raise ValueError(f"Unknown shard {shard}")
    if db_name in get_ids(client.config.databases):
        raise ValueError("database already exists")
    # MongoDB doesn't have a 'create database' command, so insert an
    #  item into a collection and then drop the collection.
    client[db_name].foo.insert({'foo': 1})
    client[db_name].foo.drop()
    if client[db_name].collection_names():
        raise ValueError("database has collections")
    primary = client['config'].databases.find_one(db_name)['primary']
    if primary != shard:
        res = client.admin.command('movePrimary', value=db_name, to=shard)
        if not res.get('ok'):
            raise RuntimeError(str(res))
    return (
        f"Successfully created {db_name} in {shard} via {client.nodes} "
        f"from {hostname}"
    )



import time
import re
import argparse

from jaraco.mongodb import helper
from .compat import query_or_command


def is_index_op(op):
    cmd = query_or_command(op) or {}
    return 'createIndexes' in cmd


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('db')
    return parser.parse_args()


def run():
    db = helper.connect_db(get_args().db)

    while True:
        ops = db.current_op()['inprog']
        index_op = next(filter(is_index_op, ops), None)
        if not index_op:
            print("No index operations in progress")
            break
        msg = index_op['msg']
        name = query_or_command(index_op)['indexes'][0]['name']
        pat = re.compile(r'Index Build( \(background\))?')
        msg = pat.sub(name, msg, count=1)
        print(msg, end='\r')
        time.sleep(5)


__name__ == '__main__' and run()



r"""
Backslash-escape text such that it is safe for MongoDB fields,
honoring `Restrictions on Field Names
<https://docs.mongodb.com/v3.4/reference/limits/#Restrictions-on-Field-Names>`_.

>>> decode(encode('my text with dots...'))
'my text with dots...'

>>> decode(encode(r'my text with both \. and literal .'))
'my text with both \\. and literal .'

>>> decode(encode('$leading dollar'))
'$leading dollar'
"""

import re


def encode(text):
    text = text.replace('\\', '\\\\')
    text = re.sub(r'^\$', '\\$', text)
    return text.replace('.', '\\D')


def unescape(match):
    char = match.group(1)
    return '.' if char == 'D' else char


def decode(encoded):
    return re.sub(r'\\(.)', unescape, encoded)



import random

import pytest


collect_ignore = [
    'jaraco/mongodb/pmxbot.py',
    # disable move-gridfs check, as it causes output capturing
    # to be disabled. See pytest-dev/pytest#3752.
    'jaraco/mongodb/move-gridfs.py',
]


@pytest.fixture(scope='function')
def database(request, mongodb_instance):
    """
    Return a clean MongoDB database suitable for testing.
    """
    db_name = request.node.name.replace('.', '_')
    database = mongodb_instance.get_connection()[db_name]
    yield database
    database.client.drop_database(db_name)


@pytest.fixture()
def bulky_collection(database):
    """
    Generate a semi-bulky collection with a few dozen random
    documents.
    """
    coll = database.bulky
    for _id in range(100):
        doc = dict(_id=_id, val=random.randint(1, 100))
        coll.insert_one(doc)
    return coll



import re


def all_databases(client, exclude=['local']):
    """
    Yield all databases except excluded (default
    excludes 'local').
    """
    return (
        client[db_name]
        for db_name in client.list_database_names()
        if db_name not in exclude
    )


def all_collections(db):
    """
    Yield all non-sytem collections in db.
    """
    include_pattern = r'(?!system\.)'
    return (
        db[name]
        for name in db.list_collection_names()
        if re.match(include_pattern, name)
    )


def purge_collection(coll):
    coll.delete_all({})


def safe_purge_collection(coll):
    """
    Cannot remove documents from capped collections
    in later versions of MongoDB, so drop the
    collection instead.
    """
    op = drop_collection if coll.options().get('capped', False) else purge_collection
    return op(coll)


def drop_collection(coll):
    coll.database.drop_collection(coll.name)


def purge_all_databases(client, op=drop_collection):
    collections = (coll for db in all_databases(client) for coll in all_collections(db))

    for coll in collections:
        op(coll)



from jaraco.mongodb import manage


def test_purge_all_databases(mongodb_instance):
    client = mongodb_instance.get_connection()
    client.test_db.test_coll.insert_one({'a': 1})
    client.test_db2.test_coll.insert_one({'b': 2})
    manage.purge_all_databases(client)
    indexes = {'system.indexes'}
    assert set(client.test_db.list_collection_names()) <= indexes
    assert set(client.test_db2.list_collection_names()) <= indexes



import os
import sys
import tempfile
import subprocess
import glob
import collections
import importlib
import shutil
import functools
import logging
import datetime
import pathlib
import contextlib

from typing import Dict, Any

import portend
from jaraco.services import paths
from jaraco import services
from tempora import timing
from . import manage
from . import cli
from . import install


log = logging.getLogger(__name__)


class MongoDBFinder(paths.PathFinder):
    windows_installed = glob.glob('/Program Files/MongoDB/Server/???/bin')
    windows_paths = [
        # symlink Server/current to Server/X.X
        '/Program Files/MongoDB/Server/current/bin',
        # symlink MongoDB to mongodb-win32-x86_64-2008plus-X.X.X-rcX
        '/Program Files/MongoDB/bin',
    ] + list(reversed(windows_installed))
    heuristic_paths = [
        # on the path
        '',
        # 10gen Debian package
        '/usr/bin',
        # custom install in /opt
        '/opt/mongodb/bin',
    ] + windows_paths

    # allow the environment to stipulate where mongodb must
    #  be found.
    env_paths = [
        os.path.join(os.environ[key], 'bin')
        for key in ['MONGODB_HOME']
        if key in os.environ
    ]
    candidate_paths = env_paths or heuristic_paths
    exe = 'mongod'
    args = ['--version']

    @classmethod
    def find_binary(cls):
        return os.path.join(cls.find_root(), cls.exe)

    @classmethod
    @contextlib.contextmanager
    def ensure(cls):
        try:
            yield cls.find_root()
        except RuntimeError:
            with tempfile.TemporaryDirectory() as tmp_dir:
                tmp = pathlib.Path(tmp_dir)
                root = install.install(target=tmp).joinpath('bin')
                cls.candidate_paths.append(root)
                yield root
            cls.candidate_paths.remove(root)


class MongoDBService(MongoDBFinder, services.Subprocess, services.Service):
    port = 27017

    process_kwargs: Dict[str, Any] = {}
    """
    keyword arguments to Popen to control the process creation
    """

    @services.Subprocess.PortFree()
    def start(self):
        super(MongoDBService, self).start()
        # start the daemon
        mongodb_data = os.path.join(sys.prefix, 'var', 'lib', 'mongodb')
        cmd = [
            self.find_binary(),
            '--dbpath=' + mongodb_data,
        ]
        self.process = subprocess.Popen(cmd, **self.process_kwargs)
        self.wait_for_pattern(r'waiting for connections on port (?P<port>\d+)')
        log.info('%s listening on %s', self, self.port)


class MongoDBInstance(MongoDBFinder, services.Subprocess, services.Service):
    mongod_args = (
        '--storageEngine',
        'ephemeralForTest',
    )

    process_kwargs: Dict[str, Any] = {}
    """
    keyword arguments to Popen to control the process creation
    """

    def merge_mongod_args(self, add_args):
        merged = list(self.mongod_args)

        if any(arg.startswith('--storageEngine') for arg in add_args):
            merged.remove('--storageEngine')
            merged.remove('ephemeralForTest')

        self.port, add_args[:] = cli.extract_param('port', add_args, type=int)

        merged.extend(add_args)
        self.mongod_args = merged

    def start(self):
        super(MongoDBInstance, self).start()
        if not hasattr(self, 'port') or not self.port:
            self.port = portend.find_available_local_port()
        self.data_dir = tempfile.mkdtemp()
        cmd = [
            self.find_binary(),
            '--dbpath',
            self.data_dir,
            '--port',
            str(self.port),
        ] + list(self.mongod_args)
        if hasattr(self, 'bind_ip') and '--bind_ip' not in cmd:
            cmd.extend(['--bind_ip', self.bind_ip])
        self.process = subprocess.Popen(cmd, **self.process_kwargs)
        portend.occupied('localhost', self.port, timeout=10)
        log.info(f'{self} listening on {self.port}')

    def get_connection(self):
        pymongo = importlib.import_module('pymongo')
        return pymongo.MongoClient('localhost', self.port)

    def purge_all_databases(self):
        manage.purge_all_databases(self.get_connection())

    def get_connect_hosts(self):
        return [f'localhost:{self.port}']

    def get_uri(self):
        return 'mongodb://' + ','.join(self.get_connect_hosts())

    def stop(self):
        super(MongoDBInstance, self).stop()
        shutil.rmtree(self.data_dir)
        del self.data_dir


class ExtantInstance:
    def __init__(self, uri):
        self.uri = uri

    def get_connection(self):
        pymongo = importlib.import_module('pymongo')
        return pymongo.MongoClient(self.uri)

    def get_uri(self):
        return self.uri


class MongoDBReplicaSet(MongoDBFinder, services.Service):
    replica_set_name = 'test'

    mongod_parameters = (
        '--oplogSize',
        '10',
    )

    def start(self):
        super(MongoDBReplicaSet, self).start()
        self.data_root = tempfile.mkdtemp()
        self.instances = list(map(self.start_instance, range(3)))
        # initialize the replica set
        self.instances[0].connect().admin.command(
            'replSetInitiate', self.build_config()
        )
        # wait until the replica set is initialized
        get_repl_set_status = functools.partial(
            self.instances[0].connect().admin.command, 'replSetGetStatus', 1
        )
        errors = importlib.import_module('pymongo.errors')
        log.info('Waiting for replica set to initialize')

        watch = timing.Stopwatch()
        while watch.elapsed < datetime.timedelta(minutes=5):
            try:
                res = get_repl_set_status()
                if res.get('myState') != 1:
                    continue
            except errors.OperationFailure:
                continue
            break
        else:
            raise RuntimeError("timeout waiting for replica set to start")

    def start_instance(self, number):
        port = portend.find_available_local_port()
        data_dir = os.path.join(self.data_root, repr(number))
        os.mkdir(data_dir)
        cmd = [
            self.find_binary(),
            '--dbpath',
            data_dir,
            '--port',
            str(port),
            '--replSet',
            self.replica_set_name,
        ] + list(self.mongod_parameters)
        log_file = self.get_log(number)
        process = subprocess.Popen(cmd, stdout=log_file)
        portend.occupied('localhost', port, timeout=50)
        log.info(f'{self}:{number} listening on {port}')
        return InstanceInfo(data_dir, port, process, log_file)

    def get_log(self, number):
        log_filename = os.path.join(self.data_root, f'r{number}.log')
        log_file = open(log_filename, 'a')
        return log_file

    def is_running(self):
        return hasattr(self, 'instances') and all(
            instance.process.returncode is None for instance in self.instances
        )

    def stop(self):
        super(MongoDBReplicaSet, self).stop()
        for instance in self.instances:
            if instance.process.returncode is None:
                instance.process.terminate()
                instance.process.wait()
            instance.log_file.close()
        del self.instances
        shutil.rmtree(self.data_root)

    def build_config(self):
        return dict(
            _id=self.replica_set_name,
            members=[
                dict(
                    _id=number,
                    host=f'localhost:{instance.port}',
                )
                for number, instance in enumerate(self.instances)
            ],
        )

    def get_connect_hosts(self):
        return [f'localhost:{instance.port}' for instance in self.instances]

    def get_uri(self):
        return 'mongodb://' + ','.join(self.get_connect_hosts())

    def get_connection(self):
        pymongo = importlib.import_module('pymongo')
        return pymongo.MongoClient(self.get_uri())


InstanceInfoBase = collections.namedtuple(
    'InstanceInfoBase', 'path port process log_file'
)


class InstanceInfo(InstanceInfoBase):
    def connect(self):
        pymongo = __import__('pymongo')
        rp = pymongo.ReadPreference.PRIMARY_PREFERRED
        return pymongo.MongoClient(f'localhost:{self.port}', read_preference=rp)



import pymongo


def project(*args, **kwargs):
    """
    Build a projection for MongoDB.

    Due to https://jira.mongodb.org/browse/SERVER-3156, until MongoDB 2.6,
    the values must be integers and not boolean.

    >>> project(a=True) == {'a': 1}
    True

    Once MongoDB 2.6 is released, replace use of this function with a simple
    dict.
    """
    projection = dict(*args, **kwargs)
    return {key: int(value) for key, value in projection.items()}


def compat_explain(cur):
    """
    Simulate MongoDB 3.0 explain result on prior versions.
    http://docs.mongodb.org/v3.0/reference/explain-results/
    """
    res = cur.explain()
    if 'nscannedObjects' in res:
        res['executionStats'] = dict(
            nReturned=res.pop('n'),
            totalKeysExamined=res.pop('nscanned'),
            totalDocsExamined=res.pop('nscannedObjects'),
            executionTimeMillis=res.pop('millis'),
        )
    return res


def upsert_and_fetch(coll, doc, **kwargs):
    """
    Fetch exactly one matching document or upsert
    the document if not found, returning the matching
    or upserted document.

    See https://jira.mongodb.org/browse/SERVER-28434
    describing the condition where MongoDB is uninterested in
    providing an upsert and fetch behavior.

    >>> instance = getfixture('mongodb_instance').get_connection()
    >>> coll = instance.test_upsert_and_fetch.items
    >>> doc = {'foo': 'bar'}
    >>> inserted = upsert_and_fetch(coll, doc)
    >>> inserted
    {...'foo': 'bar'...}
    >>> upsert_and_fetch(coll, doc) == inserted
    True
    """
    return coll.find_one_and_update(
        doc,
        {"$setOnInsert": doc},
        upsert=True,
        return_document=pymongo.ReturnDocument.AFTER,
        **kwargs
    )



import functools

import bson
import pytest
import jaraco.itertools

from jaraco.mongodb import oplog
from jaraco.mongodb import service


class TestReplacer:
    def test_rename_index_op_ns(self):
        """
        As an index operation references namespaces,
        when performing a rename operation, it's also important
        to rename the ns in the op itself.
        """
        op = {
            'ts': bson.Timestamp(1446495808, 3),
            'ns': 'airportlocker.system.indexes',
            'op': 'i',
            'o': {
                'ns': 'airportlocker.luggage.chunks',
                'key': {'files_id': 1, 'n': 1},
                'name': 'files_id_1_n_1',
                'unique': True,
            },
        }

        ren = oplog.Renamer.from_specs("airportlocker=airportlocker-us")
        ren(op)

        assert op['ns'] == 'airportlocker-us.system.indexes'
        assert op['o']['ns'] == 'airportlocker-us.luggage.chunks'

    def test_collection_rename_on_create_cmd(self):
        """
        Starting in MongoDB 3.2, a create collection is required
        before insert operations (apparently). Ensure that a renamed
        command in a create collection is renamed.
        """
        op = {
            'h': -4317026186822365585,
            't': 1,
            'ns': 'newdb.$cmd',
            'v': 2,
            'o': {'create': 'coll_1'},
            'ts': bson.Timestamp(1470940276, 1),
            'op': 'c',
        }

        ren = oplog.Renamer.from_specs("newdb.coll_1=newdb.coll_2")
        ren(op)

        assert op['o']['create'] == 'coll_2'


def make_replicaset(request):
    try:
        r_set = service.MongoDBReplicaSet()
        r_set.log_root = ''
        r_set.start()
        r_set.get_connection()
        request.addfinalizer(r_set.stop)
    except Exception as err:
        pytest.skip(f"MongoDB not available ({err})")
    return r_set


@pytest.fixture
def replicaset_factory(request):
    """
    Return a factory that can generate MongoDB replica sets
    """
    maker = functools.partial(make_replicaset, request)
    return jaraco.itertools.infinite_call(maker)


class TestOplogReplication:
    def test_index_deletion(self, replicaset_factory):
        """
        A delete index operation should be able to be applied to a replica
        """
        source = next(replicaset_factory).get_connection()
        dest = next(replicaset_factory).get_connection()
        source_oplog = oplog.Oplog(source.local.oplog.rs)
        before_ts = source_oplog.get_latest_ts()
        source.index_deletion_test.stuff.create_index("foo")
        for op in source_oplog.since(before_ts):
            oplog.apply(dest, op)

        id_index, foo_index = dest.index_deletion_test.stuff.list_indexes()

        after_ts = source_oplog.get_latest_ts()
        source.index_deletion_test.stuff.drop_index("foo_1")
        (delete_index_op,) = source_oplog.since(after_ts)
        print("attempting", delete_index_op)
        oplog.apply(dest, delete_index_op)
        (only_index,) = dest.index_deletion_test.stuff.list_indexes()
        assert only_index['name'] == '_id_'



"""
Helper functions to augment PyMongo
"""

import warnings

import pymongo
import gridfs


def filter_warnings():
    warnings.warn(
        "filter_warnings is deprecated and has no effect; do not call",
        DeprecationWarning,
    )


def connect(uri, factory=pymongo.MongoClient):
    """
    Use the factory to establish a connection to uri.
    """
    warnings.warn("do not use. Just call MongoClient directly.", DeprecationWarning)
    return factory(uri)


def connect_db(uri, default_db_name=None, factory=pymongo.MongoClient):
    """
    Use pymongo to parse a uri (possibly including database name) into
    a connected database object.

    This serves as a convenience function for the common use case where one
    wishes to get the Database object and is less concerned about the
    intermediate MongoClient object that pymongo creates (though the
    connection is always available as db.client).

    >>> db = connect_db(
    ...     'mongodb://mongodb.localhost/mydb?readPreference=secondary')
    >>> db.name
    'mydb'
    >>> db.client.read_preference
    Secondary(...)

    If no database is indicated in the uri, fall back to default.

    >>> db = connect_db('mongodb://mgo/', 'defaultdb')
    >>> db.name
    'defaultdb'

    The default should only apply if no db was present in the URI.

    >>> db = connect_db('mongodb://mgo/mydb', 'defaultdb')
    >>> db.name
    'mydb'
    """
    uri_p = pymongo.uri_parser.parse_uri(uri)
    client = factory(uri)
    return client.get_database(uri_p['database'] or default_db_name)


def get_collection(uri):
    return pymongo.uri_parser.parse_uri(uri)['collection']


def connect_gridfs(uri, db=None):
    """
    Construct a GridFS instance for a MongoDB URI.
    """
    return gridfs.GridFS(
        db or connect_db(uri),
        collection=get_collection(uri) or 'fs',
    )



"""
A MongoDB-backed CherryPy session store.

Although this module requires CherryPy, it does not impose the requirement
on the package, as any user of this module will already require CherryPy.

To enable these sessions, your code must call :meth:`Session.install()` and
then configure the CherryPy endpoint to use MongoDB sessions. For example::

    jaraco.mongodb.sessions.Session.install()

    session_config = {
        'sessions.on': True,
        'sessions.storage_type': 'MongoDB',
        'sessions.database': pymongo.MongoClient().database,
    }
    config = {
        '/': session_config,
    }

    cherrypy.quickstart(..., config=config)

The ``jaraco.modb`` module implements the codec interface, so may be used
to encode more complex objects in the session::

    session_config.update({
        'sessions.codec': jaraco.modb,
    })

"""

import datetime
import time
import logging
import pprint

import pymongo.errors
import cherrypy
import dateutil.tz

from . import timers
from . import compat

log = logging.getLogger(__name__)


class LockTimeout(RuntimeError):
    pass


class NullCodec(object):
    def decode(self, data):
        return data

    def encode(self, data):
        return data


class Session(cherrypy.lib.sessions.Session):
    """
    A MongoDB-backed CherryPy session store. Takes the following params:

    database: the pymongo Database object.

    collection_name: The name of the collection to use in the db.

    codec: An object with 'encode' and 'decode' methods, used to encode
    objects before saving them to MongoDB and decode them when loaded
    from MongoDB.

    lock_timeout: A timedelta or numeric seconds indicating how long
    to block acquiring a lock. If None (default), acquiring a lock
    will block indefinitely.
    """

    codec = NullCodec()
    "by default, objects are passed directly to MongoDB"

    def __init__(self, id, **kwargs):
        kwargs.setdefault('collection_name', 'sessions')
        kwargs.setdefault('lock_timeout', None)
        super(Session, self).__init__(id, **kwargs)
        self.setup_expiration()
        if isinstance(self.lock_timeout, (int, float)):
            self.lock_timeout = datetime.timedelta(seconds=self.lock_timeout)
        if not isinstance(self.lock_timeout, (datetime.timedelta, type(None))):
            msg = "Lock timeout must be numeric seconds or a timedelta instance."
            raise ValueError(msg)

    @classmethod
    def install(cls):
        """
        Add this session to the cherrypy session handlers. CherryPy looks
        for session classes in vars(cherrypy.lib.sessions) with a name
        in title-case followed by "Session".
        """
        cherrypy.lib.sessions.MongodbSession = cls

    @property
    def collection(self):
        return self.database[self.collection_name]

    def setup_expiration(self):
        """
        Use pymongo TTL index to automatically expire sessions.
        """
        self.collection.create_index(
            '_expiration_datetime',
            expireAfterSeconds=0,
        )

    def _exists(self):
        return bool(self.collection.find_one(self.id))

    def _load(self):
        filter = dict(
            _id=self.id,
            _expiration_datetime={'$exists': True},
        )
        projection = dict(_id=False)
        doc = self.collection.find_one(filter, projection)
        if not doc:
            return
        expiration_time = doc.pop('_expiration_datetime')
        doc = self.codec.decode(doc)
        return (doc, self._make_local(expiration_time))

    @staticmethod
    def _make_aware(local_datetime):
        return local_datetime.replace(tzinfo=dateutil.tz.tzlocal())

    @staticmethod
    def _make_utc(local_datetime):
        return Session._make_aware(local_datetime).astimezone(dateutil.tz.tzutc())

    @staticmethod
    def _make_local(utc_datetime):
        """
        For a naive utc_datetime, return the same time in the local timezone
        (also naive).
        """
        return (
            utc_datetime.replace(tzinfo=dateutil.tz.tzutc())
            .astimezone(dateutil.tz.tzlocal())
            .replace(tzinfo=None)
        )

    def _save(self, expiration_datetime):
        data = dict(self._data)
        data = self.codec.encode(data)
        # CherryPy defines the expiration in local time, which may be
        #  different for some hosts. Convert it to UTC before sticking
        #  it in the database.
        expiration_datetime = self._make_utc(expiration_datetime)
        data.update(
            _expiration_datetime=expiration_datetime,
            _id=self.id,
        )
        try:
            compat.save(self.collection, data)
        except pymongo.errors.InvalidDocument:
            log.warning(
                "Unable to save session:\n%s",
                pprint.pformat(data),
            )
            raise

    def _delete(self):
        self.collection.delete_one(self.id)

    def acquire_lock(self):
        """
        Acquire the lock. Blocks indefinitely until lock is available
        unless `lock_timeout` was supplied. If the lock_timeout elapses,
        raises LockTimeout.
        """
        # first ensure that a record exists for this session id
        try:
            self.collection.insert_one(dict(_id=self.id))
        except pymongo.errors.DuplicateKeyError:
            pass
        unlocked_spec = dict(_id=self.id, locked=None)
        lock_timer = (
            timers.Timer.after(self.lock_timeout)
            if self.lock_timeout
            else timers.NeverExpires()
        )
        while not lock_timer.expired():
            locked_spec = {'$set': dict(locked=datetime.datetime.utcnow())}
            res = self.collection.update_one(unlocked_spec, locked_spec)
            if res.raw_result['updatedExisting']:
                # we have the lock
                break
            time.sleep(0.1)
        else:
            raise LockTimeout(f"Timeout acquiring lock for {self.id}")
        self.locked = True

    def release_lock(self):
        record_spec = dict(_id=self.id)
        self.collection.update_one(record_spec, {'$unset': {'locked': 1}})
        # if no data was saved (no expiry), remove the record
        record_spec.update(_expiration_datetime={'$exists': False})
        self.collection.delete_one(record_spec)
        self.locked = False

    def __len__(self):
        return self.collection.count()
